import warnings
import importlib

try:
    from ._version import version as __version__
except ModuleNotFoundError:
    __version__ = None

from express import settings
from express.properties import BaseProperty
from typing import Type, Optional

# disable pymatgen warnings
warnings.filterwarnings("ignore")


class ExPrESS(object):
    """
    Exabyte Property Ex(ss)tractor, Sourcer, Serializer class.

    Args:
        parser_name (str): parser name.
        args (list): args passed to the parser.
        kwargs (dict): kwargs passed to the parser.
            espresso, vasp and nwchem parsers specific keys:
                work_dir (str): path to the working directory.
                stdout_file (str): path to the stdout file.
            structure parser specific keys:
                structure_string (str): structure string.
                structure_format (str): structure format, poscar or espresso-in, or nwchem-in.
    """

    def __init__(self, parser_name=None, *args, **kwargs):
        parser_class = self._get_parser_class(parser_name) if parser_name else None

        # Look up the parser class, if it exists
        if parser_class is not None:
            self.parser = parser_class(*args, **kwargs)
        else:
            self.parser = None

    def _get_parser_class(self, parser_name: str) -> Optional[Type[BaseProperty]]:
        """
        Returns parser class for a given parser name.

        Args:
            parser_name (str): parser name.

        Returns:
            class
        """
        # If-statement will return None for parser class if the parser is missing from the registry
        if parser_name in settings.PARSERS_REGISTRY:
            reference = settings.PARSERS_REGISTRY[parser_name]
            parser_class = self._get_class_by_reference(reference)
        else:
            parser_class = None

        return parser_class

    def _get_class_by_reference(self, reference: str) -> Type[BaseProperty]:
        """
        Returns class by reference.

        Args:
            reference (str): reference, e.g. express.parsers.apps.vasp.parser.VaspParser

        Returns:
             class
        """
        class_name = reference.split('.')[-1]
        module_name = '.'.join(reference.split('.')[:-1])
        return getattr(importlib.import_module(module_name), class_name)

    def _get_property_class(self, property_name: str) -> Type[BaseProperty]:
        """
        Returns property class for a given property name.

        Args:
            property_name (str): property name.

        Returns:
            class
        """
        reference: str = settings.PROPERTIES_MANIFEST[property_name]["reference"]
        return self._get_class_by_reference(reference)

    def property(self, property_name: str, *args, **kwargs) -> dict:
        """
        Extracts a given property and validates it against its schema.

        Args:
            property_name (str): property name.
            args (list): args passed to the underlying property method.
            kwargs (dict): kwargs passed to the underlying property method.

        Returns:
             dict
        """
        Property_Class = self._get_property_class(property_name)
        property_instance = Property_Class(property_name, self.parser, *args, **kwargs)
        return property_instance.serialize_and_validate()



from express.properties.non_scalar import NonScalarProperty


class TwoDimensionalPlotProperty(NonScalarProperty):
    """
    Base 2D-plot property class.
    """

    def __init__(self, name, parser, *args, **kwargs):
        super(TwoDimensionalPlotProperty, self).__init__(name, parser, *args, **kwargs)
        self.legend = None
        self.xDataArray = None
        self.yDataSeries = None

    def _serialize(self):
        serialized_data = {
            'name': self.name,
            'xAxis': self.manifest["defaults"]["xAxis"],
            'yAxis': self.manifest["defaults"]["yAxis"],
            'xDataArray': self.xDataArray,
            'yDataSeries': self.yDataSeries,
        }
        if self.legend:
            serialized_data.update({"legend": self.legend})
        return serialized_data



from mock import MagicMock

from tests import TestBase


class UnitTestBase(TestBase):
    """
    Base class for express unit tests.
    """

    def setUp(self):
        super(UnitTestBase, self).setUp()

    def tearDown(self):
        super(UnitTestBase, self).setUp()

    def get_mocked_parser(self, method_name, return_value):
        parser = MagicMock()
        parser.attach_mock(MagicMock(return_value=return_value), method_name)
        return parser



from tests.fixtures.espresso.references import *
from tests.integration import IntegrationTestBase
from express.parsers.apps.espresso.parser import EspressoParser


class TestEspressoParser(IntegrationTestBase):
    def setUp(self):
        super(TestEspressoParser, self).setUp()
        self.parser = EspressoParser(work_dir=self.workDir, stdout_file=self.stdoutFile)

    def tearDown(self):
        super(TestEspressoParser, self).setUp()

    def test_espresso_total_energy(self):
        self.assertAlmostEqual(self.parser.total_energy(), TOTAL_ENERGY, places=2)

    def test_espresso_fermi_energy(self):
        self.assertAlmostEqual(self.parser.fermi_energy(), FERMI_ENERGY, places=2)

    def test_espresso_nspins(self):
        self.assertEqual(self.parser.nspins(), NSPIN)

    def test_espresso_eigenvalues_at_kpoints(self):
        self.assertDeepAlmostEqual(self.parser.eigenvalues_at_kpoints()[0], EIGENVALUES_AT_KPOINTS_ZERO, places=2)

    def test_espresso_ibz_k_points(self):
        self.assertDeepAlmostEqual(self.parser.ibz_k_points(), IBZ_KPOINTS, places=2)

    def test_espresso_dos(self):
        self.assertDeepAlmostEqual(self.parser.dos(), DOS, places=2)

    def test_espresso_convergence_electronic(self):
        self.assertDeepAlmostEqual(self.parser.convergence_electronic(), CONVERGENCE_ELECTRONIC, places=2)

    def test_espresso_convergence_ionic(self):
        self.assertDeepAlmostEqual(self.parser.convergence_ionic(), CONVERGENCE_IONIC, places=2)

    def test_espresso_stress_tensor(self):
        self.assertDeepAlmostEqual(self.parser.stress_tensor(), STRESS_TENSOR, places=2)

    def test_espresso_pressure(self):
        self.assertAlmostEqual(self.parser.pressure(), PRESSURE, places=2)

    def test_espresso_total_force(self):
        self.assertAlmostEqual(self.parser.total_force(), TOTAL_FORCE, places=2)

    def test_espresso_atomic_forces(self):
        self.assertDeepAlmostEqual(self.parser.atomic_forces(), ATOMIC_FORCES, places=2)

    def test_espresso_total_energy_contributions(self):
        self.assertDeepAlmostEqual(self.parser.total_energy_contributions(), TOTAL_ENERGY_CONTRIBUTION, places=2)

    def test_espresso_phonon_dos(self):
        self.assertDeepAlmostEqual(self.parser.phonon_dos(), PHONON_DOS, places=2)

    def test_espresso_phonon_dispersion(self):
        self.assertDeepAlmostEqual(self.parser.phonon_dispersions(), PHONON_DISPERSIONS, places=2)



from tests.fixtures.vasp.references import *
from tests.integration import IntegrationTestBase
from express.parsers.apps.vasp.parser import VaspParser


class TestVaspParser(IntegrationTestBase):
    def setUp(self):
        super(TestVaspParser, self).setUp()
        self.parser = VaspParser(work_dir=self.workDir, stdout_file=self.stdoutFile)

    def tearDown(self):
        super(TestVaspParser, self).setUp()

    def test_vasp_total_energy(self):
        self.assertAlmostEqual(self.parser.total_energy(), TOTAL_ENERGY, places=2)

    def test_vasp_fermi_energy(self):
        self.assertAlmostEqual(self.parser.fermi_energy(), FERMI_ENERGY, places=2)

    def test_vasp_nspins(self):
        self.assertEqual(self.parser.nspins(), NSPIN)

    def test_vasp_eigenvalues_at_kpoints(self):
        self.assertDeepAlmostEqual(self.parser.eigenvalues_at_kpoints()[0], EIGENVALUES_AT_KPOINTS_ZERO, places=2)

    def test_vasp_ibz_k_points(self):
        self.assertDeepAlmostEqual(self.parser.ibz_k_points(), IBZ_KPOINTS, places=2)

    def test_vasp_dos(self):
        self.assertDeepAlmostEqual(self.parser.dos(), DOS)

    def test_vasp_convergence_electronic(self):
        self.assertDeepAlmostEqual(self.parser.convergence_electronic(), CONVERGENCE_ELECTRONIC, places=2)

    def test_vasp_convergence_ionic(self):
        self.assertDeepAlmostEqual(self.parser.convergence_ionic(), CONVERGENCE_IONIC, places=2)

    def test_vasp_stress_tensor(self):
        self.assertDeepAlmostEqual(self.parser.stress_tensor(), STRESS_TENSOR, places=2)

    def test_vasp_pressure(self):
        self.assertAlmostEqual(self.parser.pressure(), PRESSURE, places=2)

    def test_vasp_total_force(self):
        self.assertAlmostEqual(self.parser.total_force(), TOTAL_FORCE, places=2)

    def test_vasp_atomic_forces(self):
        self.assertDeepAlmostEqual(self.parser.atomic_forces(), ATOMIC_FORCES, places=2)

    def test_vasp_total_energy_contributions(self):
        self.assertDeepAlmostEqual(self.parser.total_energy_contributions(), TOTAL_ENERGY_CONTRIBUTION, places=2)

    def test_vasp_magnetic_moments(self):
        self.assertDeepAlmostEqual(self.parser.magnetic_moments(), MAGNETIC_MOMENTS, places=3)



from express.properties.scalar import ScalarProperty


class ReactionEnergyBarrier(ScalarProperty):
    """
    Reaction energy barrier.
    """

    def __init__(self, name, parser, *args, **kwargs):
        super(ReactionEnergyBarrier, self).__init__(name, parser, *args, **kwargs)
        energies = self.parser.reaction_energies()
        self.value = sorted(energies, key=lambda e: abs(e or 0), reverse=True)[0]



from tests.unit import UnitTestBase
from express.properties.scalar.density import Density

DENSITY = {
    "units": "g/cm^3",
    "name": "density",
    "value": 1
}


class DensityTest(UnitTestBase):
    def setUp(self):
        super(DensityTest, self).setUp()

    def tearDown(self):
        super(DensityTest, self).setUp()

    def test_density(self):
        parser = self.get_mocked_parser("density", 1)
        property_ = Density("density", parser)
        self.assertDeepAlmostEqual(property_.serialize_and_validate(), DENSITY)



from express.properties.scalar import ScalarProperty


class ElementalRatio(ScalarProperty):
    """
    Elemental ratio property class.
    """

    def __init__(self, name, parser, *args, **kwargs):
        super(ElementalRatio, self).__init__(name, parser, *args, **kwargs)
        self.value = self.parser.elemental_ratios()[kwargs["element"]]

    def _serialize(self):
        return {
            'name': self.name,
            'value': self.value,
            "element": self.kwargs["element"]
        }



from tests.unit import UnitTestBase
from express.properties.scalar.fermi_energy import FermiEnergy

FERMI_ENERGY = {
    "units": "eV",
    "name": "fermi_energy",
    "value": 1
}


class FermiEnergyTest(UnitTestBase):
    def setUp(self):
        super(FermiEnergyTest, self).setUp()

    def tearDown(self):
        super(FermiEnergyTest, self).setUp()

    def test_fermi_energy(self):
        parser = self.get_mocked_parser("fermi_energy", 1)
        property_ = FermiEnergy("fermi_energy", parser)
        self.assertDeepAlmostEqual(property_.serialize_and_validate(), FERMI_ENERGY)



from express.properties import BaseProperty
import os
import copy
from typing import Dict, Any
from abc import abstractmethod


class WorkflowProperty(BaseProperty):
    """
    Base class for workflow properties extracted in Express
    """

    def __init__(self, name: str, parser, *args, **kwargs):
        """
        Constructor for PyMLTrainAndPredictWorkflow

        Args:
            name (str): Name of the workflow
            parser (str): Parser to use with this workflow
        """
        super().__init__(name, parser, *args, **kwargs)
        self.name = name
        self.display_name = name

    @property
    def schema(self):
        return self.esse.get_schema_by_id("workflow")

    @property
    def workflow_specific_config(self) -> dict:
        return {}

    @property
    def common_config(self) -> dict:
        config = {
            "name": self.name,
            "displayName": self.display_name,
            "creator": {
                "_id": "",
                "cls": "User",
                "slug": ""
            },
            "owner": {
                "_id": "",
                "cls": "Account",
                "slug": ""
            },
            "schemaVersion": "0.2.0",
            "exabyteId": "",
            "hash": "",
            "_id": "",
        }
        return config

    def _serialize(self) -> dict:
        config = self.common_config
        config.update(self.workflow_specific_config)
        return config


class PyMLTrainAndPredictWorkflow(WorkflowProperty):
    """
    Next generation of ExabyteML. We expect workflows to have a format as follows:

    Workflow_Head_Subworkflow - Contains various units which prepare an ML job. For example, we may have the following
    units present.
        - An Assignment unit specifying whether the workflow is in Train or Predict mode (head-set-predict-status)
        - An IO unit to copy in the training set
        - A Conditional unit that specifies whether the train or predict setup is to be used
        - An IO unit to copy in the trained model if running in predict mode


    The final IO unit in the predict setup that we discuss, which copies in files needed for the workflow to function,
    is populated by this class's _create_download_from_object_storage_input function. It obtains a list of files
    the user has deemed important-enough to save in the workflow, and ensures they're around for the predict job.

    Workflow_Tail_Subworkflow - Contains the actual ML that the user wants to perform. The intention is that this is
    the more user-modifiable part of the ML feature. We might contain the following units here:
        - pyml:setup_variables_packages: A setup file that helps the user communicate the files needed in subsequent
                                         predict runs
        - pyml:data_input:read_csv:pandas: Uses Pandas to read in a CSV for use in further ML
        - pyml:pre_processing:standardization:sklearn: Scales the data such that it has mean=0 and variance=1, then
                                                       saves the scaler for use in predict workflows
        - pyml:model:multilayer_perceptron:sklearn: A multilayer perceptron being fit to a regression problem. Saves
                                                    the model to be used again in predict workflows
        - pyml:post_processing:parity_plot:matplotlib: Creates a parity plot if the workflow is in "Training" mode.
    """

    def __init__(self, name: str, parser,
                 *args,
                 work_dir: str,
                 upload_dir: str,
                 object_storage_data: Dict[str, str],
                 context_dir_relative_path: str,
                 workflow: Dict[str, Any],
                 **kwargs):
        """
        Constructor for PyMLTrainAndPredictWorkflow

        Args:
            name (str): Name of the workflow
            parser (Any): Parser to use with this workflow
            work_dir (str): The working directory for the job calling Express
            context_dir_relative_path (str): Relative path, from the job's working dir, to the context directory
            object_storage_data (dict): Information about the object storage provider being usec for file I/O
            workflow (dict): The workflow used to run the job
        """
        super().__init__(name, parser, *args, **kwargs)
        self.work_dir = work_dir
        self.upload_dir = upload_dir
        self.object_storage_data = object_storage_data
        self.context_dir_relative_path = context_dir_relative_path
        self.workflow = copy.deepcopy(workflow)
        self.display_name = f"[Predict Workflow] {self.workflow['name']}"

    def _create_download_from_object_storage_input(self, basename: str) -> dict:
        """
        Generates an input for a download-from-object-storage unit

        Args:
            basename (str): The basename to copy

        Returns:
            The input for a download-from-object-storage io unit

        """
        object_storage_data = copy.deepcopy(self.object_storage_data)

        # Create path name based on whether files have a relative path
        if self.context_dir_relative_path:
            path_name = (self.upload_dir, self.context_dir_relative_path, basename)
        else:
            path_name = (self.upload_dir, basename)
        object_storage_data.update({"NAME": os.path.join(*path_name)})

        io_unit_input = {
            "basename": basename,
            "pathname": self.context_dir_relative_path,
            "overwrite": False,
            "objectData": object_storage_data
        }
        return io_unit_input

    def set_io_unit_filenames(self, unit: dict) -> None:
        """
        Sets the filenames to copy inside of IO units

        Args:
            unit (dict): The IO unit to update

        Returns:
            None
        """
        context_dir_absolute_path = os.path.join(self.upload_dir, self.context_dir_relative_path)
        basenames_to_copy = os.listdir(context_dir_absolute_path)
        io_unit_inputs = map(self._create_download_from_object_storage_input, basenames_to_copy)
        unit["input"] = list(io_unit_inputs)

    def _construct_predict_subworkflows(self, train_subworkflows: list) -> list:
        """
        Given the set of training subworkflows, converts to the subworkflows defining the predict workflow.

        Args:
            train_subworkflows (list): "subworkflows" defined in the original workflow

        Returns:
            A list of subworkflows, which define the resultant predict workflow.

        """
        # Need to deepcopy to avoid changing the original subworkflow
        predict_subworkflows = copy.deepcopy(train_subworkflows)

        for subworkflow in predict_subworkflows:
            for unit in filter(lambda i: "tags" in i, subworkflow["units"]):
                tags = unit["tags"]

                # Note: This should be the *first* thing we check for, in case we have tags that add properties.
                # Remove properties if needed
                if "remove-all-results" in tags:
                    unit["results"] = []

                # The following tags can be placed in any order:

                # Set predict status
                if "pyml:workflow-type-setter" in tags:
                    unit["value"] = "True"

                # Set download-from-object-storage units
                if 'set-io-unit-filenames' in tags:
                    self.set_io_unit_filenames(unit)

                # Set predictors to print their predictions to the results tab during the predict phase
                if 'creates-predictions-csv-during-predict-phase' in tags:
                    unit["results"] = [{
                        "name": "file_content",
                        "basename": "predictions.csv",  # todo: We shouldn't be hardcoding this in to the flavors
                        "filetype": "csv"
                    }]

        return predict_subworkflows

    @property
    def is_using_dataset(self):
        return self.workflow.get("isUsingDataset", False)

    @property
    def workflow_specific_config(self) -> dict:
        """
        Generates the specific config for the new implementation of ExabyteML. The remainder of the config is
        generated inside of the parent Workflow class.

        Returns:
             dict
        """
        # Construct the "units" key inside the workflow. Here (and only here), "units" actually means "subworkflows,"
        # because that's what the key is called inside "workflow"
        train_subworkflow_units: list = self.workflow["units"]

        # Construct the "subworkflows" key inside the workflow
        train_subworkflows: list = self.workflow["subworkflows"]
        predict_subworkflows = self._construct_predict_subworkflows(train_subworkflows)

        specific_config = {
            "units": train_subworkflow_units,
            "subworkflows": predict_subworkflows,
            "isUsingDataset": self.is_using_dataset,
        }

        return specific_config


class ExabyteMLPredictWorkflow(WorkflowProperty):
    """
    Legacy implementation of Exabyte ML's predict Workflow property class.
    """

    def __init__(self, name, parser, *args, **kwargs):
        super().__init__(name, parser, *args, **kwargs)

        self.model = self.parser.model
        self.targets = self.parser.targets
        self.features = self.parser.features
        self.scaling_params_per_feature = self.parser.scaling_params_per_feature

    @property
    def workflow_specific_config(self) -> dict:
        """
        Generates the specific config for a legacy ExabyteML workflow. The remainder of the config is generated
        inside of the parent Worfklow class.

        Returns:
             dict
        """
        specific_config = {
            "units": [
                {
                    "_id": "LCthJ6E2QabYCZqf4",
                    "name": "ml_predict_subworkflow",
                    "type": "subworkflow",
                    "flowchartId": "subworkflow",
                    "head": True
                }
            ],
            "subworkflows": [
                {
                    "name": "ml_predict_subworkflow",
                    "isDraft": True,
                    "application": {
                        "version": "0.2.0",
                        "summary": "Exabyte Machine Learning Engine",
                        "name": "exabyteml",
                        "shortName": "ml",
                        "build": "Default"
                    },
                    "units": [
                        {
                            "status": "idle",
                            "statusTrack": [],
                            "head": True,
                            "flowchartId": "io",
                            "name": "input",
                            "application": {
                                "version": "0.2.0",
                                "summary": "Exabyte Machine Learning Engine",
                                "name": "exabyteml",
                                "shortName": "ml",
                                "build": "Default"
                            },
                            "results": [],
                            "next": "data_transformation_manipulation",
                            "source": "api",
                            "postProcessors": [],
                            "preProcessors": [],
                            "subtype": "dataFrame",
                            "input": [
                                {
                                    "endpoint": "dataframe",
                                    "endpoint_options": {
                                        "headers": {},
                                        "data": {
                                            "features": self.features,
                                            "ids": [],
                                            "targets": self.targets
                                        },
                                        "method": "POST",
                                        "params": {},
                                        "jobId": ""
                                    }
                                }
                            ],
                            "type": "io",
                            "monitors": []
                        },
                        {
                            "status": "idle",
                            "statusTrack": [],
                            "head": False,
                            "flowchartId": "data_transformation_manipulation",
                            "name": "clean data",
                            "monitors": [],
                            "results": [],
                            "next": "data_transformation_scale_and_reduce",
                            "application": {
                                "version": "0.2.0",
                                "summary": "Exabyte Machine Learning Engine",
                                "name": "exabyteml",
                                "shortName": "ml",
                                "build": "Default"
                            },
                            "postProcessors": [],
                            "preProcessors": [],
                            "operationType": "manipulation",
                            "operation": "data_transformation",
                            "type": "processing",
                            "inputData": {
                                "cleanMissingData": True,
                                "replaceNoneValuesWith": 0,
                                "removeDuplicateRows": True
                            }
                        },
                        {
                            "status": "idle",
                            "statusTrack": [],
                            "head": False,
                            "flowchartId": "data_transformation_scale_and_reduce",
                            "name": "scale and reduce",
                            "monitors": [],
                            "results": [],
                            "next": "score",
                            "application": {
                                "version": "0.2.0",
                                "build": "Default",
                                "name": "exabyteml",
                                "shortName": "ml",
                                "summary": "Exabyte Machine Learning Engine"
                            },
                            "postProcessors": [],
                            "preProcessors": [],
                            "operationType": "scale_and_reduce",
                            "operation": "data_transformation",
                            "type": "processing",
                            "inputData": {
                                "scaler": "standard_scaler",
                                "perFeature": self.scaling_params_per_feature,
                            }
                        },
                        {
                            "status": "idle",
                            "statusTrack": [],
                            "executable": {
                                "name": "score"
                            },
                            "flowchartId": "score",
                            "name": "score",
                            "head": False,
                            "results": [
                                {
                                    "name": "predicted_properties"
                                }
                            ],
                            "application": {
                                "version": "0.2.0",
                                "build": "Default",
                                "name": "exabyteml",
                                "shortName": "ml",
                                "summary": "Exabyte Machine Learning Engine"
                            },
                            "postProcessors": [],
                            "preProcessors": [],
                            "context": {},
                            "input": [],
                            "flavor": {
                                "name": "score"
                            },
                            "type": "execution",
                            "monitors": [
                                {
                                    "name": "standard_output"
                                }
                            ]
                        }
                    ],
                    "model": self.model,
                    "_id": "LCthJ6E2QabYCZqf4",
                    "properties": self.targets
                }
            ],
            "properties": self.targets
        }
        return specific_config



"""
Reference values for the nwchem test calculations within ExPrESS
All nwchem output values are in hartrees. ExPrESS converts units to eV.
All reference energies are in eV.
"""
TOTAL_ENERGY = -2079.18666382721904

TOTAL_ENERGY_CONTRIBUTION = {
    "one_electron": {
        "name": "one_electron",
        "value": -3350.531714067630674
    },
    "coulomb": {
        "name": "coulomb",
        "value": 1275.68347728573713
    },
    "exchange_correlation": {
        "name": "exchange_correlation",
        "value": -254.54658374762781
    },
    "nuclear_repulsion": {
        "name": "nuclear_repulsion",
        "value": 250.20815670232923
    }
}

BASIS = {
    "units": "angstrom",
    "elements": [
        {
            "id": 1,
            "value": "O"
        },
        {
            "id": 2,
            "value": "H"
        },
        {
            "id": 3,
            "value": "H"
        }
    ],
    "coordinates": [
        {
            "id": 1,
            "value": [
                0.00000000,
                0.00000000,
                0.22143053
            ]
        },
        {
            "id": 2,
            "value": [
                0.00000000,
                1.43042809,
                -0.88572213
            ]
        },
        {
            "id": 3,
            "value": [
                0.00000000,
                -1.43042809,
                -0.88572213
            ]
        }
    ]
}



from express.properties.non_scalar.two_dimensional_plot import TwoDimensionalPlotProperty


class ReactionEnergyProfile(TwoDimensionalPlotProperty):
    """
    Reaction energy profile.
    """

    def __init__(self, name, parser, *args, **kwargs):
        super(ReactionEnergyProfile, self).__init__(name, parser, *args, **kwargs)
        self.xDataArray = self.parser.reaction_coordinates()
        self.yDataSeries = [self.parser.reaction_energies()]



import numpy as np
import pymatgen as mg

from abc import abstractmethod


class ElectronicDataMixin(object):
    """
    Defines electronic interfaces.

    Note:
        THE FORMAT OF DATA STRUCTURE RETURNED MUST BE PRESERVED IN IMPLEMENTATION.
    """

    @abstractmethod
    def total_energy(self):
        """
        Returns total energy.

        Units: eV

        Returns:
             float

        Example:
             -19.00890332
        """
        pass

    @abstractmethod
    def fermi_energy(self):
        """
        Returns fermi energy.

        Units: eV

        Returns:
             float

        Example:
             6.6078556811104292
        """
        pass

    @abstractmethod
    def nspins(self):
        """
        Returns the number of spins.

        Returns:
             int

        Example:
             2
        """
        pass

    @abstractmethod
    def eigenvalues_at_kpoints(self):
        """
        Returns eigenvalues for all kpoints.

        Units:
            energy: eV
            kpoint coordinate: crystal

        Returns:
             list[dict]

        Example:
            [
                {
                    'kpoint': [-0.5, 0.5, 0.5],
                    'weight': 9.5238095E-002,
                    'eigenvalues': [
                        {
                            'energies': [-1.4498446E-001, ..., 4.6507387E-001],
                            'occupations': [1, ... , 0],
                            'spin': 0.5
                        }
                    ]
                },
                ...
            ]
        """
        pass

    @abstractmethod
    def dos(self):
        """
        Returns density of states.

        Units:
            energy: eV
            partial: eV
            total: eV

        Returns:
            dict

        Example:
            {
                "energy": [
                    -6.005000114440918,
                    -5.954999923706055,
                    -5.90500020980835
                ],
                "partial": [
                    [
                        1.6499999980444308E-17,
                        1.3080000562020133E-16,
                        7.899999954541818E-16
                    ]
                ],
                "partial_info": [
                    {
                        "electronicState": "2py",
                        "element": "Si"
                    },
                    {
                        "electronicState": "2px",
                        "element": "Si"
                    },
                    {
                        "electronicState": "1s",
                        "element": "Si"
                    },
                    {
                        "electronicState": "2pz",
                        "element": "Si"
                    }
                ],
                "total": [
                    0.00012799999967683107,
                    0.0010100000072270632,
                    0.006130000110715628
                ]
            }
        """
        pass

    @abstractmethod
    def convergence_electronic(self):
        """
        Extracts convergence electronic.

        Returns:
             list[float]

        Example:
            [
                1.4018213061907816,
                0.5939946985677435,
                0.007003124785903934,
                0.0010198831091687887,
                4.2041606287774244e-05,
                7.619190783544846e-06
            ]
        """
        pass

    @abstractmethod
    def total_energy_contributions(self):
        """
        Extracts total energy contributions.

        Units:
            energy: eV

        Returns:
            dict

        Example:
            {
                "harrisFoulkes": {
                    "name": "harris_foulkes",
                    "value": -258.6293887585482
                },
                "ewald": {
                    "name": "ewald",
                    "value": -226.94126871332813
                },
                "oneElectron": {
                    "name": "one_electron",
                    "value": 68.65366986552296
                },
                "smearing": {
                    "name": "smearing",
                    "value": -0.0
                },
                "hartree": {
                    "name": "hartree",
                    "value": 17.72349166363712
                },
                "exchangeCorrelation": {
                    "name": "exchange_correlation",
                    "value": -118.06528742483022
                }
            }
        }
        """
        pass

    @abstractmethod
    def reaction_energies(self):
        """
        Returns reaction energies.

        Units: eV

        Returns:
             list[float]

        Example:
             [
                0.0000000000
                0.0336646452
                0.1282974628
                0.2032892090
                0.1282940160
                0.0336634943
                0.0000000063
            ]
        """
        pass

    @abstractmethod
    def reaction_coordinates(self):
        """
        Returns reaction coordinates.

        Returns:
             list[float]

        Example:
             [
                0.0000000000
                0.1932752934
                0.3596145525
                0.5000030773
                0.6403901111
                0.8067278352
                1.0000000000
            ]
        """
        pass

    def reaction_coordinates_from_poscars(self, poscars):
        """
        Returns reaction coordinates based on the given poscars.

        See `NEBAnalysis` class in http://pymatgen.org/_modules/pymatgen/analysis/transition_state.html for more info.

        Args:
            poscars (list): structures in POSCAR format.

        Returns:
             list
        """
        structures = [mg.Structure.from_str(poscar, "poscar") for poscar in poscars]
        prev = structures[0]
        reaction_coordinates = [0]
        for structure in structures[1:]:
            dists = np.array([s2.distance(s1) for s1, s2 in zip(prev, structure)])
            reaction_coordinates.append(np.sqrt(np.sum(dists ** 2)))
            prev = structure
        reaction_coordinates = np.cumsum(reaction_coordinates)
        return ((1 / reaction_coordinates[-1]) * reaction_coordinates).tolist()



from express.properties.utils import to_array_with_ids
from express.properties.non_scalar import NonScalarProperty


class AtomicConstraints(NonScalarProperty):
    """
    Atomic constraints property class.
    """

    def __init__(self, name, parser, *args, **kwargs):
        super(AtomicConstraints, self).__init__(name, parser, *args, **kwargs)

    def _serialize(self):
        return {
            'name': self.name,
            "values": to_array_with_ids(self.parser.atomic_constraints())
        }



from tests.unit import UnitTestBase
from express.properties.scalar.volume import Volume

VOLUME = {
    "units": "angstrom^3",
    "name": "volume",
    "value": 1
}


class VolumeTest(UnitTestBase):
    def setUp(self):
        super(VolumeTest, self).setUp()

    def tearDown(self):
        super(VolumeTest, self).setUp()

    def test_volume(self):
        parser = self.get_mocked_parser("volume", 1)
        property_ = Volume("volume", parser)
        self.assertDeepAlmostEqual(property_.serialize_and_validate(), VOLUME)



from express.properties.scalar import ScalarProperty


class FermiEnergy(ScalarProperty):
    """
    The highest energy level occupied by electrons in a system.
    """

    def __init__(self, name, parser, *args, **kwargs):
        super(FermiEnergy, self).__init__(name, parser, *args, **kwargs)
        self.value = self.parser.fermi_energy()



from tests.unit import UnitTestBase
from tests.fixtures.data import SPACE_GROUP_SYMBOL
from express.properties.non_scalar.symmetry import Symmetry

SYMMETRY = {
    "spaceGroupSymbol": "Fd-3m",
    "tolerance": {
        "value": 0.3,
        "units": "angstrom"
    },
    "name": "symmetry"
}


class SymmetryTest(UnitTestBase):
    def setUp(self):
        super(SymmetryTest, self).setUp()

    def tearDown(self):
        super(SymmetryTest, self).setUp()

    def test_symmetry(self):
        parser = self.get_mocked_parser("space_group_symbol", SPACE_GROUP_SYMBOL)
        property_ = Symmetry("symmetry", parser)
        self.assertDeepAlmostEqual(property_.serialize_and_validate(), SYMMETRY)





import os
import re
import string
import numpy as np
import xml.etree.ElementTree as ET

from express.parsers.settings import Constant
from express.parsers.settings import GENERAL_REGEX
from express.parsers.formats.xml import BaseXMLParser

TAG_VALUE_CAST_MAP = {
    'character': lambda v, s, c: v,
    'integer': lambda v, s, c: np.array([int(_) for _ in re.findall(GENERAL_REGEX.int_number, v)]).reshape([s // c, c]),
    'real': lambda v, s, c: np.array([float(_) for _ in re.findall(GENERAL_REGEX.double_number, v)]).reshape(
        [s // c, c]),
    'logical': lambda v, s, c: False if 'F' in v else True

}


class EspressoXMLParser(BaseXMLParser):
    """
    Espresso XML parser class.

    Args:
        xml_file_path (str): path to the xml file.
    """

    def __init__(self, xml_file_path):
        super(EspressoXMLParser, self).__init__(xml_file_path)

    def _get_xml_tag_value(self, tag):
        """
        This function helps casting xml tag value to the the type defined in the tag attribute. It xml tag's text
        holds a string, vector, (vector, matrix, constant) of either int or float.

        Args:
            tag (xml.etree.ElementTree.Element): An Element instance of ElementTree XML class.

        Returns:
            str | float | int | bool | ndarray: depending on the tag type attribute.
        """
        type = tag.attrib.get("type")
        size = int(tag.attrib.get("size", 1))
        columns = int(tag.attrib.get("columns", 1))
        result = TAG_VALUE_CAST_MAP[type](tag.text, size, columns)
        return result[0][0] if size == 1 and type not in ['logical', 'character'] else result

    def fermi_energy(self):
        """
        Extracts fermi energy.

        Returns:
            float
        """
        bs_tag = self.root.find("BAND_STRUCTURE_INFO")
        return self._get_xml_tag_value(bs_tag.find("FERMI_ENERGY")) * Constant.HARTREE

    def nspins(self):
        """
        Extracts the number of number of spin components.

        Returns:
             int
        """
        bs_tag = self.root.find("BAND_STRUCTURE_INFO")
        return self._get_xml_tag_value(bs_tag.find("NUMBER_OF_SPIN_COMPONENTS"))

    def final_lattice_vectors(self, reciprocal=False):
        """
        Extracts lattice.

        Args:
            reciprocal (bool): whether to extract reciprocal lattice.

        Returns:
            dict

        Examples:
            {
                'vectors': {
                    'a': [-0.561154473, -0.000000000, 0.561154473],
                    'b': [-0.000000000, 0.561154473, 0.561154473],
                    'c': [-0.561154473, 0.561154473, 0.000000000],
                    'alat': 9.44858082
                }
             }
        """
        vector_tag = 'a'
        lattice_tag = 'DIRECT_LATTICE_VECTORS'
        units_tag = 'UNITS_FOR_DIRECT_LATTICE_VECTORS'

        if reciprocal:
            vector_tag = 'b'
            lattice_tag = 'RECIPROCAL_LATTICE_VECTORS'
            units_tag = 'UNITS_FOR_RECIPROCAL_LATTICE_VECTORS'

        vectors = {}
        cell_tag = self.root.find("CELL")
        lattice_units_tag = cell_tag.find(lattice_tag).find(units_tag)
        for vector in cell_tag.find(lattice_tag):
            if vector.tag.startswith(vector_tag):
                vectors.update({
                    string.ascii_lowercase[int(vector.tag[1]) - 1]: (
                        (Constant.BOHR if not reciprocal else 1.0) * self._get_xml_tag_value(vector)[0]).tolist()
                })
        vectors.update({'alat': 1.0})
        return {'vectors': vectors, 'units': 'angstrom'} if not reciprocal else {'vectors': vectors}

    def get_inverse_reciprocal_lattice_vectors(self):
        """
        Returns inverse reciprocal lattice vectors to convert cartesian (2pi/a) point to crystal.
        """
        reciprocal_lattice = self.final_lattice_vectors(reciprocal=True)
        lattice_array = [reciprocal_lattice['vectors'][i] for i in ['a', 'b', 'c']]
        return np.linalg.inv(np.array(lattice_array))

    def eigenvalues_at_kpoints(self):
        """
        Returns eigenvalues for all kpoints.

        Returns:
             list

        Example:
            [
                {
                    'kpoint': [-0.5, 0.5, 0.5],
                    'weight': 9.5238095E-002,
                    'eigenvalues': [
                        {
                            'energies': [-1.4498446E-001, ..., 4.6507387E-001],
                            'occupations': [1, ... , 0],
                            'spin': 0.5
                        }
                    ]
                },
                ...
            ]
        """
        eigenvalues_at_kpoints = []
        for eigenvalue_tag in self.root.find("EIGENVALUES"):
            cartesianKPoint = self._get_xml_tag_value(eigenvalue_tag.find("K-POINT_COORDS"))[0]
            crystalKPoint = np.dot(cartesianKPoint, self.get_inverse_reciprocal_lattice_vectors())
            eigenvalues_at_kpoint = {
                "kpoint": crystalKPoint.tolist(),
                "weight": self._get_xml_tag_value(eigenvalue_tag.find("WEIGHT")),
                "eigenvalues": []
            }
            for datafile_tag in [t for t in eigenvalue_tag.iter() if t.tag.startswith('DATAFILE')]:
                eigenval_file = os.path.join(self.xml_dir_name, datafile_tag.attrib.get("iotk_link"))
                energies, occupations = self._parse_eigenvalue_file(eigenval_file)
                eigenvalues_at_kpoint['eigenvalues'].append({
                    'energies': (np.array(energies) * Constant.HARTREE).tolist(),
                    'occupations': occupations,
                    'spin': 0.5 if datafile_tag.tag in ['DATAFILE', 'DATAFILE.1'] else -0.5
                })
            eigenvalues_at_kpoints.append(eigenvalues_at_kpoint)
        return eigenvalues_at_kpoints

    def _parse_eigenvalue_file(self, eigenval_xml_path):
        """
        Extracts eigenvalues from a given file.

        Args:
            eigenval_xml_path (str): path to eigenvalue xml file.

        Returns:
            tuple: (energies, occupations)
        """
        root = ET.parse(eigenval_xml_path).getroot()
        energies = [float(_) for _ in re.findall(GENERAL_REGEX.double_number, root.find('EIGENVALUES').text)]
        occupations = [float(_) for _ in re.findall(GENERAL_REGEX.double_number, root.find('OCCUPATIONS').text)]
        return energies, occupations

    def final_basis(self):
        """
        Extracts basis.

        Returns:
            dict

        Example:
            {
                'units': 'angstrom',
                'elements': [{'id': 1, 'value': 'Si'}, {'id': 2, 'value': 'Si'}],
                'coordinates': [{'id': 1, 'value': [0.0, 0.0, 0.0]}, {'id': 2, 'value': [0.0, 0.0, 0.0]}]
             }
        """
        elements, coordinates = [], []
        ion_tag = self.root.find("IONS")
        for atom in ion_tag:
            if atom.tag.startswith("ATOM"):
                elements.append({
                    'id': int(atom.tag[5:]),
                    'value': atom.attrib.get("SPECIES").strip(' \t\n\r')
                })
                coordinates.append({
                    'id': int(atom.tag[5:]),
                    'value': (Constant.BOHR * np.array(atom.attrib.get("tau").split()).astype(np.float)).tolist()
                })

        return {
            'units': 'angstrom',
            'elements': elements,
            'coordinates': coordinates
        }



from express.parsers.settings import GENERAL_REGEX

NEB_DIR_PREFIX = "0"
NEB_STD_OUT_FILE = "stdout"
XML_DATA_FILE = "vasprun.xml"

_COMMON_REGEX = r"{0}\s+({1})\s+({1})\s+({1})\s+({1})\s+({1})\s+({1})"

REGEX = {
    "total_energy": {
        "regex": r"F=\s+({0})".format(GENERAL_REGEX.double_number),
        "start_flag": "entering main loop",
        "occurrences": -1,
        "output_type": "float",
    },
    "ibz_kpoints": {
        "regex": r'({double})\s+({double})\s+({double})\s+{double}'.format(double=GENERAL_REGEX.double_number)
    },
    "convergence_electronic": {
        "regex": r".+:\s+(\d+)\s+{0}\s+({0})".format(GENERAL_REGEX.double_number),
        "output_type": "float",
        "match_groups": [1, 2]
    },
    "ion_positions_block": {
        "regex": r"POSITION.+?-{5,}\n(.+?)-{5,}"
    },
    "lattice_vectors": {
        "regex": r"direct lattice vectors.+\n"
                 r"\s+({double})\s+({double})\s+({double}).+\n"
                 r"\s+({double})\s+({double})\s+({double}).+\n"
                 r"\s+({double})\s+({double})\s+({double})".format(double=GENERAL_REGEX.double_number)
    },
    "basis_vectors": {
        "regex": r'\s+({double})\s+({double})\s+({double})\s+{double}\s+{double}\s+{double}'.format(
            double=GENERAL_REGEX.double_number)
    },
    "pressure": {
        "regex": r"external pressure\s+=\s+({0})\s+kB".format(GENERAL_REGEX.double_number),
        "occurrences": -1,
        "output_type": "float",
    },
    "total_force": {
        "regex": r"total drift:\s+({0})\s+({0})\s+({0})".format(GENERAL_REGEX.double_number),
        "occurrences": -1,
        "output_type": "float",
        "match_groups": [1, 2, 3]
    },
    "zero_point_energy": {
        "regex": r"f\s\s=.*2PiTHz\s+\d+\.\d+\s+cm\-1\s+({0})\s+meV".format(GENERAL_REGEX.double_number),
        "start_flag": "Eigenvectors and eigenvalues of the dynamical matrix",
        "output_type": "float",
    }
}

TOTAL_ENERGY_CONTRIBUTIONS = {
    "hartree": {
        "regex": _COMMON_REGEX.format("Hartree", GENERAL_REGEX.double_number),
        "occurrences": -1,
        "output_type": "float",
        "match_groups": [1, 2, 3, 4, 5, 6]
    },
    "ewald": {
        "regex": _COMMON_REGEX.format("Ewald", GENERAL_REGEX.double_number),
        "occurrences": -1,
        "output_type": "float",
        "match_groups": [1, 2, 3, 4, 5, 6]
    },
    "exchange_correlation": {
        "regex": _COMMON_REGEX.format("E\(xc\)", GENERAL_REGEX.double_number),
        "occurrences": -1,
        "output_type": "float",
        "match_groups": [1, 2, 3, 4, 5, 6]
    }
}



import os
from express.parsers.settings import ASE_FORMATS

def find_file(name, path):
    """
    Finds file with the specified name in the given path.

    Args:
        name (str): file name.
        path (str): starting path for search.

    Returns:
        str: absolute file path (if found)
    """
    for root, dirs, files in os.walk(path, followlinks=True):
        for file in files:
            if name in file:
                return os.path.join(root, file)


def find_files_by_name_substring(name, path):
    matches = []
    for root, dirs, files in os.walk(path, followlinks=True):
        for file_ in files:
            if name in file_:
                matches.append(os.path.join(root, file_))
    return matches


def get_element_counts(basis):
    """
    Returns chemical elements with their count wrt their original order in the basis.
    Note: entries for the same element separated by another element are considered separately.
    [{"count":1, "value":"Zr"}, {"count":23, "value":"H"}, {"count":11, "value":"Zr"}, {"count":1, "value":"H"}]
    """
    element_counts = []
    previous_element = None
    for index, element in enumerate(basis["elements"]):
        if previous_element and previous_element["value"] == element["value"]:
            element_counts[-1]["count"] += 1
        else:
            element_counts.append({
                "count": 1,
                "value": element["value"]
            })
        previous_element = basis["elements"][index]
    return element_counts


def lattice_basis_to_poscar(lattice, basis, basis_units="cartesian"):
    element_counts = get_element_counts(basis)
    return "\n".join([
        "material",
        "1.0",
        "\t".join(["{0:14.9f}".format(x) for x in lattice["vectors"]["a"]]),
        "\t".join(["{0:14.9f}".format(x) for x in lattice["vectors"]["b"]]),
        "\t".join(["{0:14.9f}".format(x) for x in lattice["vectors"]["c"]]),
        " ".join((e["value"] for e in element_counts)),
        " ".join((str(e["count"]) for e in element_counts)),
        basis_units,
        "\n".join([" ".join(["{0:14.9f}".format(v) for v in x["value"]]) for x in basis["coordinates"]])
    ])

def convert_to_ase_format(format):
    """
    Function converts the format keywords used in this code to their
    corresponding ase keywords based on a predisposed dictionary of values.

    Returns:
        Str

    Example:
        format="poscar" --> format="vasp"
    """
    ase_format = ASE_FORMATS[format]
    return ase_format



from express.properties.non_scalar.two_dimensional_plot import TwoDimensionalPlotProperty


class DensityOfStates(TwoDimensionalPlotProperty):
    """
    Describes the number of electronic states per interval of energy at each energy level that are available to be
    occupied. There are also projections of total electronic density onto each of the atomic states that are often
    useful.
    """

    def __init__(self, name, parser, *args, **kwargs):
        super(DensityOfStates, self).__init__(name, parser, *args, **kwargs)
        dos = self.parser.dos()
        self.xDataArray = [dos['energy']]
        self.yDataSeries = [dos['total']]
        self.legend = [{}] + dos['partial_info']
        self.yDataSeries.extend([pdos for pdos in dos['partial']])



ZERO_WEIGHT_KPOINT_THRESHOLD = 1e-7

SCALAR_PROPERTIES_MANIFEST = {
    'total_energy': {
        'reference': 'express.properties.scalar.total_energy.TotalEnergy'
    },
    'fermi_energy': {
        'reference': 'express.properties.scalar.fermi_energy.FermiEnergy'
    },
    'pressure': {
        'reference': 'express.properties.scalar.pressure.Pressure'
    },
    'total_force': {
        'reference': 'express.properties.scalar.total_force.TotalForce'
    },
    'volume': {
        'reference': 'express.properties.scalar.volume.Volume'
    },
    'density': {
        'reference': 'express.properties.scalar.density.Density'
    },
    'elemental_ratio': {
        'reference': 'express.properties.scalar.elemental_ratio.ElementalRatio'
    },
    'p-norm': {
        'reference': 'express.properties.scalar.p_norm.PNorm'
    },
    'zero_point_energy': {
        'reference': 'express.properties.scalar.zero_point_energy.ZeroPointEnergy'
    },
    'surface_energy': {
        'reference': 'express.properties.scalar.surface_energy.SurfaceEnergy'
    },
    'reaction_energy_barrier': {
        'reference': 'express.properties.scalar.reaction_energy_barrier.ReactionEnergyBarrier'
    },
}

NON_SCALAR_PROPERTIES_MANIFEST = {
    'band_gaps': {
        'reference': 'express.properties.non_scalar.bandgaps.BandGaps'
    },
    'density_of_states': {
        'reference': 'express.properties.non_scalar.two_dimensional_plot.density_of_states.DensityOfStates'
    },
    'band_structure': {
        'reference': 'express.properties.non_scalar.two_dimensional_plot.band_structure.BandStructure'
    },
    'stress_tensor': {
        'reference': 'express.properties.non_scalar.stress_tensor.StressTensor'
    },
    'atomic_forces': {
        'reference': 'express.properties.non_scalar.atomic_forces.AtomicForces'
    },
    'atomic_constraints': {
        'reference': 'express.properties.non_scalar.atomic_constraints.AtomicConstraints'
    },
    'total_energy_contributions': {
        'reference': 'express.properties.non_scalar.total_energy_contributions.TotalEnergyContributions'
    },
    'material': {
        'reference': 'express.properties.material.Material'
    },
    'symmetry': {
        'reference': 'express.properties.non_scalar.symmetry.Symmetry'
    },
    'workflow:ml_predict': {
        'reference': 'express.properties.workflow.ExabyteMLPredictWorkflow'
    },
    'workflow:pyml_predict': {
       'reference': 'express.properties.workflow.PyMLTrainAndPredictWorkflow'
    },
    'phonon_dos': {
        'reference': 'express.properties.non_scalar.two_dimensional_plot.phonon_dos.PhononDOS'
    },
    'phonon_dispersions': {
        'reference': 'express.properties.non_scalar.two_dimensional_plot.phonon_dispersions.PhononDispersions'
    },
    'magnetic_moments': {
        'reference': 'express.properties.non_scalar.magnetic_moments.MagneticMoments'
    },
    'reaction_energy_profile': {
        'reference': 'express.properties.non_scalar.two_dimensional_plot.reaction_energy_profile.ReactionEnergyProfile'
    },
    'potential_profile': {
        'reference': 'express.properties.non_scalar.two_dimensional_plot.potential_profile.PotentialProfile'
    },
    'charge_density_profile': {
        'reference': 'express.properties.non_scalar.two_dimensional_plot.charge_density_profile.ChargeDensityProfile'
    },
    'file_content': {
        'reference': 'express.properties.non_scalar.file_content.FileContent'
    },
}

CONVERGENCE_PROPERTIES = {
    'convergence_electronic': {
        'reference': 'express.properties.convergence.electronic.ConvergenceElectronic'
    },
    'convergence_ionic': {
        'reference': 'express.properties.convergence.ionic.ConvergenceIonic'
    }
}

PROPERTIES_MANIFEST = dict()
PROPERTIES_MANIFEST.update(CONVERGENCE_PROPERTIES)
PROPERTIES_MANIFEST.update(SCALAR_PROPERTIES_MANIFEST)
PROPERTIES_MANIFEST.update(NON_SCALAR_PROPERTIES_MANIFEST)

PARSERS_REGISTRY = {
    'espresso': 'express.parsers.apps.espresso.parser.EspressoParser',
    'vasp': 'express.parsers.apps.vasp.parser.VaspParser',
    'nwchem': 'express.parsers.apps.nwchem.parser.NwchemParser',
    'structure': 'express.parsers.structure.StructureParser',
    'exabyteml': 'express.parsers.exabyteml.ExabyteMLParser'
}

PRECISION = 4



from mock import MagicMock

from tests.unit import UnitTestBase
from express.properties.non_scalar.bandgaps import BandGaps
from tests.fixtures.data import EIGENVALUES_AT_KPOINTS, IBZ_K_POINTS

BAND_GAPS = {
    "eigenvalues": [
        {
            "eigenvalues": [
                {
                    "energies": [
                        6.2693,
                        6.2693,
                        8.7114,
                        8.7114
                    ],
                    "spin": 0.5,
                    "occupations": [
                        1.0,
                        1.0,
                        0.0,
                        0.0
                    ]
                }
            ],
            "weight": 0.25,
            "kpoint": [
                0,
                0,
                0
            ]
        },
        {
            "eigenvalues": [
                {
                    "energies": [
                        5.0608,
                        5.0609,
                        7.695,
                        9.4927
                    ],
                    "spin": 0.5,
                    "occupations": [
                        1.0,
                        1.0,
                        0.0,
                        0.0
                    ]
                }
            ],
            "weight": 0.5,
            "kpoint": [
                0.2887,
                0.2041,
                -0.5
            ]
        },
        {
            "eigenvalues": [
                {
                    "energies": [
                        5.0608,
                        5.0609,
                        7.695,
                        9.4927
                    ],
                    "spin": 0.5,
                    "occupations": [
                        1.0,
                        1.0,
                        0.0,
                        0.0
                    ]
                }
            ],
            "weight": 0.25,
            "kpoint": [
                0.0,
                -0.6124,
                0.0
            ]
        },
        {
            "eigenvalues": [
                {
                    "energies": [
                        3.4107,
                        3.4107,
                        6.9196,
                        6.9196
                    ],
                    "spin": 0.5,
                    "occupations": [
                        1.0,
                        1.0,
                        0.0,
                        0.0
                    ]
                }
            ],
            "weight": 0.5,
            "kpoint": [
                0.2887,
                -0.4082,
                -0.5
            ]
        },
        {
            "eigenvalues": [
                {
                    "energies": [
                        5.0609,
                        5.0609,
                        7.695,
                        9.4927
                    ],
                    "spin": 0.5,
                    "occupations": [
                        1.0,
                        1.0,
                        0.0,
                        0.0
                    ]
                }
            ],
            "weight": 0.25,
            "kpoint": [
                -0.5774,
                0.2041,
                0.0
            ]
        },
        {
            "eigenvalues": [
                {
                    "energies": [
                        3.4107,
                        3.4107,
                        6.9196,
                        6.9196
                    ],
                    "spin": 0.5,
                    "occupations": [
                        1.0,
                        1.0,
                        0.0,
                        0.0
                    ]
                }
            ],
            "weight": 0.25,
            "kpoint": [
                -0.5774,
                -0.4082,
                0.0
            ]
        }
    ],
    "values": [
        {
            "units": "eV",
            "kpointConduction": [
                0.0,
                0.0,
                0.0
            ],
            "type": "direct",
            "value": 2.4420081600000003,
            "kpointValence": [
                0.0,
                0.0,
                0.0
            ]
        },
        {
            "units": "eV",
            "kpointConduction": [
                -4.8471013318887174e-17,
                -0.4999999999999998,
                -0.5000000000000001
            ],
            "type": "indirect",
            "value": 0.65023092000000027,
            "kpointValence": [
                0.0,
                0.0,
                0.0
            ]
        }
    ],
    "name": "band_gaps"
}


class BandGapsTest(UnitTestBase):
    def setUp(self):
        super(BandGapsTest, self).setUp()

    def tearDown(self):
        super(BandGapsTest, self).setUp()

    def test_band_gaps(self):
        parser = MagicMock()
        parser.attach_mock(MagicMock(return_value=1), "nspins")
        parser.attach_mock(MagicMock(return_value=6.6), "fermi_energy")
        parser.attach_mock(MagicMock(return_value=None), "band_gaps_direct")
        parser.attach_mock(MagicMock(return_value=None), "band_gaps_indirect")
        parser.attach_mock(MagicMock(return_value=IBZ_K_POINTS), "ibz_k_points")
        parser.attach_mock(MagicMock(return_value=EIGENVALUES_AT_KPOINTS), "eigenvalues_at_kpoints")
        property_ = BandGaps("band_gaps", parser)
        self.assertDeepAlmostEqual(property_.serialize_and_validate(), BAND_GAPS)



from abc import abstractmethod


class ExabyteMLDataMixin(object):
    """
    Defines Exabyte ML interfaces.

    Note:
        THE FORMAT OF DATA STRUCTURE RETURNED MUST BE PRESERVED IN IMPLEMENTATION.
    """

    @abstractmethod
    def data_per_property(self):
        pass

    @abstractmethod
    def precision_per_property(self):
        pass

    @abstractmethod
    def scaling_params_per_feature(self):
        pass

    @abstractmethod
    def band_gaps_direct(self):
        pass

    @abstractmethod
    def band_gaps_indirect(self):
        pass



from esse import ESSE
from abc import abstractmethod


class BaseProperty(object):
    """
    Base Property class.

    Args:
        name (str): property name.
        parser: an instance of parser class.
        args (list): property-specific args.
        kwargs (dict): property-specific kwargs.
    """

    def __init__(self, name, parser, *args, **kwargs):
        self.name, self.parser = name, parser
        self.args, self.kwargs = args, kwargs
        self.esse = ESSE()
        self.manifest = self.esse.get_property_manifest(self.name)

    @abstractmethod
    def _serialize(self):
        pass

    @property
    def schema(self):
        return self.esse.get_schema_by_id(self.manifest["schemaId"])

    def serialize_and_validate(self):
        """
        Serialize the property and validates it against the schema.

        Returns:
            dict
        """
        instance = self._serialize()
        self.esse.validate(instance, self.schema)
        return instance

    def safely_invoke_parser_method(self, method_name, *args, **kwargs):
        if hasattr(self.parser, method_name):
            return getattr(self.parser, method_name)(*args, **kwargs)



import re
import builtins


class BaseTXTParser(object):
    """
    Base text parser class.
    Args:
        work_dir (str): path to the working directory.
    """

    def __init__(self, work_dir):
        self.work_dir = work_dir

    def _general_output_parser(self, text, regex, output_type, start_flag=None, occurrences=0, match_groups=[]):
        """
        General function for extracting data from a text output. It extracts basic values using regex patterns. Based
        on the input regex pattern, this function uses re.findall method to find every instance of the pattern inside
        the text.
        Args:
            text (str): text to search
            regex (str): regex pattern.
            output_type (str): output type.
            start_flag (str): a symbol in the output file to be used as the starting point (for speedup and accuracy).
            occurrences (int): number of desired line counts to be processed. If negative, last occurrences is extracted:
                                - N < 0: extract the last N instance(s). Forms a list if |N| > 1
                                - N = 0: extract all of of the occurred instances and forms a list.
                                - N > 0: extract the first N instance(s). Forms a list if |N| > 1
            match_groups (list): list of match groups to be used for the regex.
        Return:
            any
        """
        start_index = text.rfind(start_flag) if start_flag else 0
        pattern = re.compile(regex, re.I | re.MULTILINE)
        cast = getattr(builtins, output_type)
        # output type depends on the number of values required. List or single number.
        result = [] if len(match_groups) > 1 or abs(occurrences) > 1 or occurrences == 0 else None

        match = pattern.findall(text[start_index:])
        if match:
            occurrences = len(match) if occurrences == 0 else occurrences
            match = match[occurrences:] if occurrences < 0 else match[:occurrences]
            if isinstance(result, list):
                for m in match:
                    result.append([cast(m[i - 1]) for i in match_groups]) if match_groups else result.append(cast(m))
            else:
                result = cast(match[0][0]) if isinstance(match[0], tuple) else cast(match[0])
        return result



from express.properties.utils import to_array_with_ids
from express.properties.non_scalar import NonScalarProperty


class MagneticMoments(NonScalarProperty):

    def __init__(self, name, parser, *args, **kwargs):
        super(MagneticMoments, self).__init__(name, parser, *args, **kwargs)

    def _serialize(self):
        return {
            'name': self.name,
            "units": self.manifest["defaults"]["units"],
            "values": to_array_with_ids(self.parser.magnetic_moments())
        }



import os

from tests import TestBase


class IntegrationTestBase(TestBase):
    """
    Base class for express integration tests.
    """

    def setUp(self):
        super(IntegrationTestBase, self).setUp()

    def tearDown(self):
        super(IntegrationTestBase, self).setUp()

    @property
    def workDir(self):
        return os.path.join(self.rootDir, self.getManifest()["workDir"])

    @property
    def stdoutFile(self):
        return os.path.join(self.rootDir, self.getManifest()["stdoutFile"])



import io
import pymatgen as mg
from ase.io import read, write

from express.parsers.structure import StructureParser


class CrystalParser(StructureParser):
    """
    Crystal  parser class.

    Args:
        args (list): args passed to the parser.
        kwargs (dict): kwargs passed to the parser.
            structure_string (str): structure string.
            structure_format (str): structure format, poscar or espresso-in.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


    def volume(self):
        """
        Returns volume.

        Reference:
            func: express.parsers.mixins.ionic.IonicDataMixin.volume
        """
        return self.structure.volume

    def density(self):
        """
        Returns density.

        Reference:
            func: express.parsers.mixins.ionic.IonicDataMixin.density
        """
        return self.structure.density
